#!/usr/bin/env python

import re
import sys
#import codecs
from collections import defaultdict

leaves = re.compile(r"[^( ]+:\d+")
leaf_nt = re.compile(r"\([^( ]+:\d+")
root = re.compile(r"^(?:\()[^( ):]+")
alignments = re.compile(r"([^( ]+):(\d+)((@[0123456789]+)*)")

ltr = {}

def reduce_tag(tag):
    return tag.split("-")[0]
    
def x_tag(tag):
    #return tag
    return "X"

tag_dispatcher = {"re": reduce_tag, "x": x_tag, "full": (lambda x: x)}

LTAG = tag_dispatcher[sys.argv[1]]
RTAG = tag_dispatcher[sys.argv[2]]

def moses_string(rule):
    
    leaf_nts = set()
    leaf_nts_rhs = set()
    new_lhs = ""
    
    lhs_mapping = {}
    lhs, rhs = rule.split(" -> ")
    lhs_leaves = [tuple(l.split(":")) for l in leaves.findall(lhs)]

    lhs_pos = [(a, int(b)) for (a, b) in lhs_leaves]
    for (i, (a, b)) in enumerate(lhs_pos):
        # map heap positions to leaf positions
        lhs_mapping[b] = (i, a)
        ltr[b] = []
    
    for nt in leaf_nt.findall(lhs):
        leaf_nts.add(int(nt.split(":")[1]))

    new_rhs = "" 
    new_aligns = ""
    lex_aligns = ""

    rhs_mapping = {}
    rhs = rhs.split(" ; ")
    rhs_roots = ""
    rhs_norm = ""
    lhsleaves_norm = ""
    rhsleaves_norm = ""
    
    for rhs_part in rhs:
        align_string = ""
        lex_align = ""
        rhs_norm_part = ""
        for nt in leaf_nt.findall(rhs_part):
            leaf_nts_rhs.add(int(nt.split(":")[1]))
        for (i, r) in enumerate(alignments.findall(rhs_part)):
            a = int(r[1])
            s = r[0]
            b = r[2:][0]

            if b:
                for x in b[1:].split("@"):
#                    print x
                    ltr[int(x)].append(a)
                    if a in leaf_nts_rhs:
                        # we need an alignment here
#                        print lhs_mapping[int(x)]
                        align_string += "%s-%s " % (lhs_mapping[int(x)][0], i)
                    else:
#                        print lhs_mapping[int(x)]
                        lex_align += "%s-%s " % (lhs_mapping[int(x)][0], i)

                rhs_mapping[a] = s

                if a in leaf_nts_rhs:
                    new_rhs += "[%s][%s] " % (LTAG(lhs_mapping[int(x)][1]), RTAG(s))
                    rhs_norm_part += "_%s" % (RTAG(s),)
                else:
                    new_rhs += "%s " % (s.lower(), )
                    rhs_norm_part += "_%s" % (s.lower(),)
            else:
                new_rhs += "%s " % (s.lower(), )
                rhs_norm_part += "_%s" % (s.lower(),)

        rhs_part_root_match = root.match(rhs_part)
        if rhs_part_root_match:
            rhs_part_root = RTAG(rhs_part_root_match.group(0)[1:])
            rhs_roots += "_%s" % (rhs_part_root,)
        else:
            rhs_roots += "_%s" % (rhs_part.split("@")[0])
            sys.stderr.write("%s\n" % (rule,))
            sys.stderr.write("%s\n" % (rhs_part.split("@")[0],))
        rhs_root = RTAG(rhs_part_root)
        new_rhs += " [%s] || " % (rhs_root,)
        new_aligns += "%s|| " % (align_string,)
        lex_aligns += "%s|| " % (lex_align,)
        
        rhsleaves_norm += "__%s" % (rhs_norm_part,)

    for (k, v) in reversed(sorted(lhs_mapping.iteritems())):
        if k in leaf_nts:
            new_lhs += "[%s]%s " % (LTAG(v[1]), 
                "".join("[%s]" % (RTAG(rhs_mapping[rhs_aligned]),) for rhs_aligned in ltr[k]))
            lhsleaves_norm += "_%s" % (LTAG(v[1]),)
        else:
            new_lhs += "%s " % (v[1].lower(),)
            lhsleaves_norm += "_%s" % (v[1].lower(),)
    
    new_root = root.match(rule).group(0)[1:]
    new_lhs += " [%s]" % (LTAG(new_root), )
    
    norm = {}
    norm["lhsroot"] = new_root
    norm["lhs"] = new_lhs
    norm["rhs"] = new_rhs[:-4]
    norm["rhsroots"] = rhs_roots
    norm["lhsleaves"] = lhsleaves_norm
    norm["rhsleaves"] = rhsleaves_norm
    norm["roots"] = new_root + "->" + rhs_roots
    
    # LHS normalization
    return (norm, ("[MBOT] ||| %s ||| %s") % (new_lhs, new_rhs[:-4],), new_aligns[:-4], lex_aligns[:-4])
        
    
if __name__ == "__main__":
    normcounts = defaultdict(lambda: defaultdict(int))
    norms = {}
    counts = defaultdict(int)
#    for normkey in sys.argv[3:]:
#        normcounts[normkey] = defaultdict(int)
    for (ii, line) in enumerate(sys.stdin):
        if (ii % 10000 == 0):
            sys.stderr.write('%s\r' % (ii,))
        sys.stderr.write("%s\t " % (str(ii)))
        sys.stderr.write(line)

        line = line.strip() 
        if line:
            #print line
            x = moses_string(line)
            try:
                (norm, rule, alignment, lex_alignment) = x
#                sys.stderr.write(str(norm))
#                sys.stderr.write("\n")
            except(ValueError):
                sys.stderr.write("Faulty rule: %s!\n" % (x,))
                continue
            counts[(rule, alignment, lex_alignment)] += 1
            norms[(rule, alignment, lex_alignment)] = norm
            for (normkey, normvalue) in norm.iteritems():
                normcounts[normkey][normvalue] += 1
    sys.stderr.write('Done counting and computing weights, writing Moses strings now.\n')
        
                
    for (ii, ((rule, alignment, lex_alignment), count)) in enumerate(counts.iteritems()):
        if (ii % 100000 == 0):
            sys.stderr.write('%s\r' % (ii,))
        weights = []
        rule_norm = norms[(rule, alignment, lex_alignment)]
        for normkey in sys.argv[3:]:
            weights.append(float(count) / normcounts[normkey][rule_norm[normkey]])
        weights = " ".join(str(w) for w in weights)
        lex_plus = " || ".join(("%s %s" % (a.strip(), b.strip())).strip()
            for (a, b) in zip(alignment.split("||"), lex_alignment.split("||")))
        sys.stdout.write("%s ||| %s 1 1 2.178 ||| %s ||| %s\n" % (rule, weights, lex_plus, count))

