#include <sstream>
#include <fstream>
#include <iostream>
#include <vector>
#include <iterator>
#include <algorithm>
#include <string>
#include <utility>
#include <map>
#include <set>
#include <math.h>

using namespace std;

typedef pair<string,string> lexKey;
typedef map<lexKey,double> LexMap;

int DEBUG=0;
float PENALTY = 0.01;
int THRESHOLD = 10;

 std::vector< std::set<int> > alignedToT;
 std::vector< std::set<int> > alignedToS;
 std::vector<int> targetOffset;

//routines
void put_into_map( ifstream * ifs, LexMap * map );
double get_from_map(pair<string,string> key, LexMap * map);
std::string buildAlignmentVectors(string alignString,std::vector< std::set<int> > alignedToT, std::vector< std::set<int> > alignedToS);
vector<string> buildTargetPhrases(string targetString);
double computeLexicalTranslationForSource(vector<string> sourcePhrase, vector<string> targetPhrase, std::vector< std::set<int> > alignedToT, LexMap * map);
double computeLexicalTranslationForTarget(vector<string> sourcePhrase, vector<string> targetPhrase, std::vector< std::set<int> > alignedToS, LexMap * map);
std::vector<std::string> tokenize(const std::string& str, const std::string& delimiters = " \t");
std::vector<std::string> tokenizeMultiCharSeparator(const std::string& str, const std::string& separator);
bool CheckIfTerminal(string token);
void Discount(int frequency);

void put_into_map( std::ifstream * ifs, LexMap * map )
{
  string line;
  while(getline( *ifs, line ))
  {
	if (DEBUG>4) std::cout << "READING LINE " << line << std::endl;
	vector<string> tokenizedLine = tokenize(line, " ");
	if(tokenizedLine.size() != 3)
	{
		std::cout << "A line should contain 3 token" << std::endl;
		abort();
	}
	const string sourceWord = tokenizedLine[0];
	if (DEBUG>4) std::cout << "Source word " << sourceWord << std::endl;
	const string targetWord = tokenizedLine[1];
	if (DEBUG>4) std::cout << "Target word " << targetWord << std::endl;
	string proba = tokenizedLine[2];

	stringstream ss(proba);
	double prob = 0;
	ss >> prob;
	pair<string,string> myKey = make_pair(sourceWord,targetWord);
	map->insert(make_pair(myKey,prob));
  }
}

double get_from_map(pair<string,string> key, LexMap * map)
{
  if (DEBUG>2) std::cout << "looking for " << key.first << " " << key.second << std::endl;
   LexMap::iterator itr_map;
   itr_map=map->find(key);

  if(itr_map != map->end())
  return{itr_map->second};

  else
  {
	if (DEBUG>4) std::cout << "Pair not found : penalty !!!" << std::endl;
	return PENALTY;
  }
}

std::string buildAlignmentVectors(string alignString, vector<string> sourcePhrase, vector<string> targetPhrase)
{
  alignedToT.clear();
  alignedToS.clear();
  std::string ret = " ";

  int i = 0;
  for(i=0; i<sourcePhrase.size(); i++)
    {
      std::set<int> s;
      alignedToS.push_back(s);
      if (DEBUG>4) std::cout << "SRC: " << i << " " << sourcePhrase[i] << std::endl;
    }
  for(i=0; i<targetPhrase.size(); i++)
    {
      std::set<int> s;
      alignedToT.push_back(s);
      if (DEBUG>4) std::cout << "TGT: " << i << " " << targetPhrase[i] << std::endl;
    }
	vector<string> alignStringFields;
	vector<string> :: iterator itr_align_fields;
	alignStringFields = tokenizeMultiCharSeparator(alignString , "||");

	//transform the alignments to get source and target words
	int alignCounter = 0;
	for(itr_align_fields = alignStringFields.begin(); itr_align_fields != alignStringFields.end(); itr_align_fields++)
    	{
		if(itr_align_fields != alignStringFields.begin())
		{ret+="|| ";}
        string alignField = *itr_align_fields;
        vector<string> allAlignments = tokenize(alignField, " ");
		vector<string> :: iterator itr_singleAlign;

		for(itr_singleAlign = allAlignments.begin(); itr_singleAlign != allAlignments.end(); itr_singleAlign++)
		{
			string singleAlignment = *itr_singleAlign;
			std::vector<string> alignment = tokenize(singleAlignment,"-");
			string sourceAlign = alignment[0];
			string targetAlign = alignment[1];
			int alignSrc;
			int alignTgt;
			istringstream (sourceAlign) >> alignSrc;
			istringstream (targetAlign) >> alignTgt;
			if(targetPhrase[alignTgt + targetOffset[alignCounter]].at(0)!='[' || sourcePhrase[alignSrc].at(0)!='[')
			{
				//we have a terminal
				if (DEBUG>2)std::cout << "SOURCEALIGN: " << sourceAlign << " " << sourcePhrase[alignSrc] << std::endl;
				if (DEBUG>2)std::cout << "TARGETALIGN: " << targetAlign << " " << targetPhrase[alignTgt + targetOffset[alignCounter]] << std::endl;

			    alignedToT[alignTgt + targetOffset[alignCounter]].insert(alignSrc);
			    alignedToS[alignSrc].insert(alignTgt + targetOffset[alignCounter]);

			  }
			else
			{
				ret+=singleAlignment;
				ret+=" ";
			}
		}
		alignCounter++;
        }
	return ret;
}

vector<string> buildTargetPhrases(string targetString)
{
        targetOffset.clear();
        int offset=0;
	vector<string> targetFields;
	vector<string> :: iterator itr_targetFields;
	targetFields = tokenizeMultiCharSeparator(targetString , "||" );
	vector<string> allTargets;
	for(itr_targetFields = targetFields.begin(); itr_targetFields != targetFields.end(); itr_targetFields++)
	{
    targetOffset.push_back(offset);
		vector<string> targetToken = tokenize(*itr_targetFields," ");
		offset += targetToken.size();
		vector<string> :: iterator itr_target_tok;
		for(itr_target_tok = targetToken.begin(); itr_target_tok != targetToken.end(); itr_target_tok++)
		{
			allTargets.push_back(*itr_target_tok);
		}
	}
return allTargets;
}

double computeLexicalTranslationForTarget(vector<string> sourcePhrase, vector<string> targetPhrase, LexMap * map)
{
  if(DEBUG > 4) {std::cout << "COMPUTING LEXICAL FOR TARGET"<< std::endl;}
  // lexical translation probability
  double lexScore = 1.0;
  // all target words have to be explained
  for(int ti=0; ti< alignedToT.size(); ti++) {
    if (targetPhrase[ti].at(0)!='[')
      {
        const set< int >  srcIndices = alignedToT[ ti ];
        if (srcIndices.empty()) {
          // explain unaligned word by NULL
          const string myTarget = targetPhrase[ti];
          const string null = "NULL";
          lexScore *= get_from_map(make_pair(null,myTarget),map);
          if(DEBUG > 4){std::cout << "Lexical Score for NULL and target : " << myTarget << " : "<< get_from_map(make_pair(null,myTarget),map) << std::endl;}
        } else {
          // go through all the aligned words to compute average
          double thisWordScore = 0;
          for (set< int >::const_iterator p(srcIndices.begin()); p != srcIndices.end(); ++p) {
            thisWordScore += get_from_map(make_pair(targetPhrase[ ti ],sourcePhrase[*p]),map);
            lexScore *= thisWordScore / (double)srcIndices.size();
            if(DEBUG > 4){std::cout << "Lexical Score for target : " << targetPhrase[ ti ] << " and source "<< sourcePhrase[*p] << " : "<< get_from_map(make_pair(targetPhrase[ ti ],sourcePhrase[*p]),map) << std::endl;}
          }
        }
      }
  }
  return lexScore;
}

double computeLexicalTranslationForSource(vector<string> sourcePhrase, vector<string> targetPhrase, LexMap * map)
{
	if(DEBUG>4){std::cout << "COMPUTING LEXICAL FOR SOURCE"<< std::endl;}
  // lexical translation probability
  double lexScore = 1.0;
  // all target words have to be explained
  for(int si=0; si< alignedToS.size(); si++) {
    if (sourcePhrase[si].at(0)!='[')
    {
      const set< int > & targetIndices = alignedToS[ si ];
      if (targetIndices.empty()) {
        const string mySource = sourcePhrase[si];
        const string null = "NULL";
        lexScore *= get_from_map(make_pair(mySource,null),map);
        if(DEBUG>4){std::cout << "Lexical Score for source : " << mySource << " and NULL : "<< get_from_map(make_pair(mySource,null),map) << std::endl;}
      } else {
        // go through all the aligned words to compute average
        double thisWordScore = 0;
        for (set< int >::const_iterator p(targetIndices.begin()); p != targetIndices.end(); ++p) {
          thisWordScore += get_from_map(make_pair(sourcePhrase[si],targetPhrase[ *p ]), map);
          if(DEBUG>4){std::cout << "Lexical Score for source : " << sourcePhrase[si] << " and target : "<< targetPhrase[ *p ] << ": " << get_from_map(make_pair(sourcePhrase[si],targetPhrase[ *p ]), map) << std::endl;}
        }
        lexScore *= thisWordScore / (double)targetIndices.size();
      }
    }
  }
  return lexScore;
}

std::vector<std::string> tokenize(const std::string& str,
    const std::string& delimiters)
{
  std::vector<std::string> tokens;
  // Skip delimiters at beginning.
  size_t lastPos = str.find_first_not_of(delimiters, 0);
  // Find first "non-delimiter".
  size_t pos     = str.find_first_of(delimiters, lastPos);

  while (std::string::npos != pos || std::string::npos != lastPos) {
    // Found a token, add it to the vector.
    tokens.push_back(str.substr(lastPos, pos - lastPos));
    // Skip delimiters.  Note the "not_of"
    lastPos = str.find_first_not_of(delimiters, pos);
    // Find next "non-delimiter"
    pos = str.find_first_of(delimiters, lastPos);
  }

  return tokens;
}

std::vector<std::string> tokenizeMultiCharSeparator(
  const std::string& str,
  const std::string& separator)
{
  std::vector<std::string> tokens;

  int pos = 0;
  // Find first "non-delimiter".
  size_t nextPos     = str.find(separator, pos);

  while (nextPos != std::string::npos) {
    // Found a token, add it to the vector.
    tokens.push_back(str.substr(pos, nextPos - pos));
    // Skip delimiters.  Note the "not_of"
    pos = nextPos + separator.size();
    // Find next "non-delimiter"
    nextPos	= str.find(separator, pos);
  }
  tokens.push_back(str.substr(pos, nextPos - pos));

  return tokens;
}

float Discount(float transWeight, int frequency)
{
	float discountedFrequency = transWeight;
	if(frequency < THRESHOLD)
	{
	  discountedFrequency = transWeight * PENALTY;
	}
	return discountedFrequency;
}

int main(int argc, char* args[])
{
  //load map
  LexMap mySourceMap;
  LexMap myTargetMap;

  std::ifstream ifs1( args[1] );
  put_into_map( &ifs1, &mySourceMap );

  std::ifstream ifs2( args[2] );
  put_into_map( &ifs2, &myTargetMap );

  std::ifstream ifRules( args[3] );

  string line;

  while(getline( ifRules , line ) && line.size()>3)
  {
    if (DEBUG>3)std::cerr << "A line : "  << line << std::endl;
    vector<string> tokens;
    tokens = tokenizeMultiCharSeparator( line , "|||" );
    string sourcePhraseString = tokens[1];
    string targetPhraseString = tokens[2];
    string scoreString = tokens[3];
    string alignString = tokens[4];
    string frequency = tokens[5];

    vector<string> targetToken = buildTargetPhrases(targetPhraseString);
    vector<string> sourceToken = tokenize(sourcePhraseString, " ");
    std::string newAlign = buildAlignmentVectors(alignString, sourceToken, targetToken);
    //lexical translation for source (indirect)
    double scoreSRC = computeLexicalTranslationForSource(sourceToken, targetToken, &mySourceMap);
    //lexical translation for target (direct)
    double scoreTGT = computeLexicalTranslationForTarget(sourceToken, targetToken, &myTargetMap);
    
    //Split Score string in order to obtain all scores
    vector<string> tokenizedLine = tokenize(scoreString, " ");

    bool isBackwards = 0;
    
    if(tokenizedLine.size() == 5)
      {
        isBackwards = 1;
      }

    string transProb = tokenizedLine[0];

    //discounting
    int frequInt = atoi(frequency.c_str());
    float transProbFlt = atof(transProb.c_str());
    float newTransProb = Discount(transProbFlt,frequInt);


    if(isBackwards)
      {
        string backwardsTransProb = tokenizedLine[1];
        std::cout <<  "[MBOT] ||| " << sourcePhraseString << "|||" << targetPhraseString << "||| " << backwardsTransProb << " " 
                  << scoreSRC << " " << newTransProb << " " << scoreTGT <<  " 2.718 |||" << newAlign << "|||" << frequency << std::endl;
      }
    else
      {
        std::cout <<  "[MBOT] ||| " << sourcePhraseString << "|||" << targetPhraseString << "||| " << newTransProb << " " << scoreSRC 
                  <<  " "  << scoreTGT << " 2.718 |||" << newAlign << "|||" << frequency << std::endl;
      }
  }
}



